import inspect
import os
import subprocess
from itertools import cycle

import numpy as np

from scm.plams.interfaces.adfsuite.scmjob import SCMJob, SCMResults
from scm.plams.tools.units import Units

__all__ = ["CRSResults", "CRSJob"]


class CRSResults(SCMResults):
    """A |SCMResults| subclass for accessing results of |CRSJob|."""

    _kfext = ".crskf"
    _rename_map = {"CRSKF": "$JN.crskf"}

    @property
    def section(self) -> str:
        try:  # Return the cached value if possible
            return self._section
        except AttributeError:
            try:
                self._section = self.job.settings.input.property._h.upper()
            except AttributeError:
                self._section = self.job.settings.input.t.upper()

            return self._section

    def get_energy(self, energy_type: str = "deltag", compound_idx: int = 0, unit: str = "kcal/mol") -> float:
        """Returns the solute solvation energy from an Activity Coefficients calculation."""
        E = self.readkf(self.section, energy_type)[compound_idx]
        return Units.convert(E, "kcal/mol", unit)

    def get_activity_coefficient(self, compound_idx: int = 0) -> float:
        """Return the solute activity coefficient from an Activity Coefficients calculation."""
        return self.readkf(self.section, "gamma")[compound_idx]

    def get_sigma_profile(self, subsection: str = "profil", as_df: bool = False) -> dict:
        r"""Grab all sigma profiles, returning a dictionary of Numpy Arrays.

        Values of :math:`\sigma` are stored under the ``"σ (e/A**2)"`` key.

        Results can be returned as a Pandas DataFrame by settings *as_df* to ``True``.

        The returned results can be plotted by passing them to the :meth:`CRSResults.plot` method.

        .. note::
            *as_df* = ``True`` requires the Pandas_ package.
            Plotting requires the `matplotlib <https://matplotlib.org/index.html>`__ package.

        .. _Pandas: https://pandas.pydata.org/

        """
        args = (subsection, "σ (e/A**2)", "chdval")
        try:
            return self._get_array_dict("SIGMAPROFILE", *args, as_df=as_df)
        except KeyError:
            return self._get_array_dict("PURESIGMAPROFILE", *args, as_df=as_df)

    def get_sigma_potential(self, subsection: str = "mu", unit: str = "kcal/mol", as_df: bool = False) -> dict:
        r"""Grab all sigma profiles, expressed in *unit*, and return a dictionary of Numpy Arrays.

        Values of :math:`\sigma` are stored under the ``"σ (e/A**2)"`` key.

        Results can be returned as a Pandas DataFrame by settings *as_df* to ``True``.

        The returned results can be plotted by passing them to the :meth:`CRSResults.plot` method.

        .. note::
            *as_df* = ``True`` requires the Pandas_ package.
            Plotting requires the `matplotlib <https://matplotlib.org/index.html>`__ package.

        .. _Pandas: https://pandas.pydata.org/

        """
        args = (subsection, "σ (e/A**2)", "chdval")
        try:
            return self._get_array_dict("SIGMAPOTENTIAL", *args, unit=unit, as_df=as_df)
        except KeyError:
            return self._get_array_dict("PURESIGMAPOTENTIAL", *args, unit=unit, as_df=as_df)

    def get_prop_names(self, section=None) -> list:
        r"""Read the section of the .crskf file and return a list of the properties that were calculated.  The section argument can be supplied to look at previously-calculated results.  If no section name is supplied, the function defaults to using the most recent property that was calculated."""
        if section is None:
            section = self.section
        try:
            return self._kf.get_skeleton()[section]
        except KeyError:
            raise KeyError("Cannot find section name: " + str(section))

    def get_results(self, section=None) -> dict:
        r"""Read the section from the most recent calculation type and return the result as a dictionary."""
        if section is None:
            section = self.section

        output = getattr(self, "_prop_dict", False)
        if output and output["section"] == section:
            return output

        props = self.get_prop_names()
        try:
            props.remove("ncomp")
            props.remove("nitems")
        except ValueError:
            raise ValueError("Results object is missing or incomplete.")

        # first get the two ranges for the indices
        ncomp = self.readkf(section, "ncomp")
        nitems = self.readkf(section, "nitems")
        try:
            nstruct = self.readkf(section, "nstruct")
        except:
            nstruct = ncomp

        np_dict = {"section": section}
        np_dict["ncomp"] = ncomp
        for prop in props:
            tmp = self.readkf(section, prop)
            if (prop == "filename") or (prop == "name"):
                chunk_length = len(tmp) // ncomp
                np_dict[prop] = [tmp[i : i + chunk_length].strip() for i in range(0, len(tmp), chunk_length)]
                continue
            if prop == "struct names":
                chunk_length = len(tmp) // nstruct
                np_dict[prop] = [tmp[i : i + chunk_length].strip() for i in range(0, len(tmp), chunk_length)]
                continue
            if not isinstance(tmp, list):
                np_dict[prop] = tmp
            else:
                np_dict[prop] = np.array(tmp)
                if len(tmp) == ncomp * nitems:
                    np_dict[prop].shape = (ncomp, nitems)

        setattr(self, "_prop_dict", np_dict)
        return np_dict

    def get_multispecies_dist(self):
        """
        This function returns multispecies distribution for each (compound,structure) pair.  The format is a list
        with indices corresponding to compound indices.  Each item in the list is a dictionary with a structure name : list pair, where the structure name corresponds to a structure the compound can be exist as and the list is the distribution of that compound in that structure over the number of points (mole fractions, temperatures, pressures).
        """
        res = self.get_results()
        property_name = res["property"].rstrip()

        if property_name == "LOGP":
            nPhase = 2
        else:
            nPhase = 1

        ncomp = self.readkf(self.section, "ncomp")
        struct_names = res["struct names"]
        num_points = self.readkf(self.section, "nitems")
        valid_structs = [[] for _ in range(ncomp)]
        comp_dist = res["comp distribution"].flatten()
        for i in range(len(struct_names)):
            for j in range(ncomp):
                if res["valid structs"][i * ncomp + j]:
                    valid_structs[j].append(struct_names[i])

        compositions = [{vs: [] for vs in valid_structs[i]} for i in range(ncomp)]
        idx = 0
        for i in range(ncomp):
            for nfrac in range(num_points):
                for k in range(nPhase):
                    for j in range(len(valid_structs[i])):
                        compositions[i][valid_structs[i][j]].append(comp_dist[idx])
                        idx += 1

        return compositions

    def plot(self, *arrays: "np.ndarray", x_axis: str = None, plot_fig: bool = True, x_label=None, y_label=None):
        """Plot, show and return a series of COSMO-RS results as a matplotlib Figure instance.

        Accepts the output of, *e.g.*, :meth:`CRSResults.get_sigma_profile`:
        A dictionary of Numpy arrays or a Pandas DataFrame.

        Returns a matplotlib Figure_ instance which can be further modified to the users liking.
        Automatic plotting of the resulting figure can be disabled with the *plot_fig* argument.

        .. note::
            This method requires the `matplotlib <https://matplotlib.org/index.html>`__ package.

        .. note::
            The name of the dictionary/DataFrame key containing the index (*i.e.* the x-axis) can,
            and should, be manually specified in *x_axis* if a custom *x_axis* is passed
            to :meth:`CRSResults._get_array_dict`.
            This argument can be ignored otherwise.

        .. _Figure: https://matplotlib.org/api/_as_gen/matplotlib.figure.Figure.html#matplotlib.figure.Figure

        """  # noqa

        def get_x_axis(array, x_axis):
            """Find and return the index and its name."""
            if x_axis is None:
                return np.arange(array.shape[1])

            if isinstance(x_axis, str):
                ret = self._prop_dict[x_axis]
            else:
                ret = np.array(x_axis, copy=False)
            ret = ret.ravel()  # Flatten it
            return ret[: array.shape[1]]

        # Check if matplotlib is installed
        try:
            import matplotlib

            matplotlib.use("TkAgg") if plot_fig else matplotlib.use("Agg")
            import matplotlib.pyplot as plt
        except ImportError:
            method = self.__class__.__name__ + ".plot"
            raise ImportError("{}: this method requires the 'matplotlib' package".format(method))

        self.get_results()

        # Create a dictionary of 1d arrays
        array_dict = {}
        for array in arrays:
            name = None
            if isinstance(array, str):  # Array refers to a section in the kf file
                name = array
                array = self._prop_dict[array]

            # Ensure it's a 2D array
            array = np.array(array, ndmin=2, dtype=float, copy=False)

            # Fill the array dict with 1d arrays
            base_key = "" if name is None else name + " "
            iterator = enumerate(array, 1) if array.shape[0] != 1 else zip(cycle(" "), array)
            for i, array_1d in iterator:
                key = f"{base_key}{i}"
                array_dict[key] = array_1d
        # Retrieve the index and its name
        index = get_x_axis(array, x_axis)
        # print ("INDEX::::", index)
        if x_label is None:
            if isinstance(x_axis, str):
                x_label = x_axis
            else:
                x_label = ""

        if y_label is None:
            y_label = ""

        # Assign various series to the plot
        fig, ax = plt.subplots()
        for k, v in array_dict.items():
            ax.plot(index, v, label=k)

        # Add the legend and x-label
        ax.legend()
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)

        # Show and return
        if plot_fig:
            plt.show()
        return fig

    def _get_array_dict(
        self,
        section: str,
        subsection: str,
        x_axis: str,
        index_subsection: str,
        unit: str = "kcal/mol",
        as_df: bool = False,
    ) -> dict:
        """Create dictionary or DataFrame containing all values in *section*/*subsection*.

        Takes the following arguments:
            * The *section*/*subsection* of the desired quantity.
            * The desired name of the index (*x_axis*).
            * The name of subsection containing the index (*index_subsection*).
            * The *unit* of the output quanty (ignore this keyword if not applicable).
            * If the result should be returned as Pandas DataFrame (*as_df*).

        """
        ret = self._construct_array_dict(section, subsection, unit)

        # Create the index
        index = self.readarray(section, index_subsection, dtype=float)
        if section in ("BINMIXCOEF", "COMPOSITIONLINE", "TERNARYMIX"):
            ncomponent = 3 if section == "TERNARYMIX" else 2
            index.shape = ncomponent, len(index) // ncomponent
            iterator = np.nditer(index.astype(str), flags=["external_loop"], order="F")
            ret[x_axis] = np.array([" / ".join(i for i in item) for item in iterator])
        else:
            ret[x_axis] = index

        # Return a dictionary of arrays or a DataFrame
        if not as_df:
            return ret
        else:
            return self._dict_to_df(ret, section, x_axis)

    def _construct_array_dict(self, section: str, subsection: str, unit: str = "kcal/mol") -> dict:
        """Construct dictionary containing all values in *section*/*subsection*."""
        # Use filenames as keys
        _filenames = self.readkf(section, "filename").split()
        filenames = [_filenames] if not isinstance(_filenames, list) else _filenames

        # Grab the keys and the number of items per key
        keys = [os.path.basename(key) for key in filenames] + ["Total"]
        nitems = self.readkf(section, "nitems")

        # Use sigma profiles/potentials as values
        ratio = Units.conversion_ratio("kcal/mol", unit)
        values = ratio * self.readarray(section, subsection, dtype=float)
        values.shape = len(values) // nitems, nitems

        ret = dict(zip(keys, values))
        try:
            ret["Total"] = self.readarray(section, subsection + "tot", dtype=float)
        except KeyError:
            pass
        return ret

    @staticmethod
    def _dict_to_df(array_dict: dict, section: str, x_axis: str):
        """Attempt to convert a dictionary into a DataFrame."""
        try:
            import pandas as pd
        except ImportError:
            method = inspect.stack()[2][3]
            raise ImportError("{}: as_df=True requires the 'pandas' package".format(method))

        index = pd.Index(array_dict.pop(x_axis), name=x_axis)
        df = pd.DataFrame(array_dict, index=index)
        df.columns.name = section.lower()
        return df


class CRSJob(SCMJob):
    """A |SCMJob| subclass intended for running COSMO-RS jobs."""

    _command = "crs"
    _result_type = CRSResults
    _subblock_end = "end"

    def __init__(self, **kwargs) -> None:
        """Initialize a :class:`CRSJob` instance."""
        super().__init__(**kwargs)
        self.settings.ignore_molecule = True

    @staticmethod
    def database() -> str:
        database_path = os.path.join(os.environ["SCM_PKG_ADFCRSDIR"], "ADFCRS-2018")
        if not os.path.isdir(database_path):
            raise FileNotFoundError("The ADFCRS-2018 database does not seem to be installed")
        return database_path

    @staticmethod
    def coskf_from_database(name: str) -> str:
        if not name.endswith(".coskf"):
            name += ".coskf"
        return os.path.join(CRSJob.database(), name)

    @staticmethod
    def cos_to_coskf(filename: str) -> str:
        """Convert a .cos file into a .coskf file with the :code:`$AMSBIN/cosmo2kf` command.

        Returns the filename of the new .coskf file.

        """
        filename_out = filename + "kf"
        try:
            amsbin = os.environ["AMSBIN"]
        except KeyError:
            raise EnvironmentError(
                "cos_to_coskf: Failed to load 'cosmo2kf' from '$AMSBIN/'; "
                "the 'AMSBIN' environment variable has not been set"
            )

        args = [os.path.join(amsbin, "cosmo2kf"), filename, filename_out]
        subprocess.run(args)
        return filename_out



import numpy as np

try:
    from scipy.spatial.distance import cdist

    scipy_present = True
except ImportError:
    scipy_present = False

from scm.plams.tools.units import Units

__all__ = [
    "rotation_matrix",
    "axis_rotation_matrix",
    "distance_array",
    "angle",
    "dihedral",
    "cell_shape",
    "cellvectors_from_shape",
]


def rotation_matrix(vec1, vec2):
    """
    Calculate the rotation matrix rotating *vec1* to *vec2*. Vectors can be any containers with 3 numerical values. They don't need to be normalized. Returns 3x3 numpy array.
    """
    vec1, vec2 = np.array(vec1), np.array(vec2)
    a = vec1 / np.linalg.norm(vec1)
    b = vec2 / np.linalg.norm(vec2)

    # avoid division by zero in case of antiparallel vectors
    if abs(1 + np.dot(a, b)) < 1e-8:
        return -np.eye(3)

    v1, v2, v3 = np.cross(a, b)
    M = np.array([[0, -v3, v2], [v3, 0, -v1], [-v2, v1, 0]])
    return np.identity(3) + M + np.dot(M, M) / (1 + np.dot(a, b))


def axis_rotation_matrix(vector, angle, unit="radian"):
    """
    Calculate the rotation matrix rotating along the *vector* by *angle* expressed in *unit*.

    *vector* can be any container with 3 numerical values. They don't need to be normalized. A positive angle denotes counterclockwise rotation, when looking along *vector*. Returns 3x3 numpy array.
    """

    vector /= np.linalg.norm(vector)
    v0, v1, v2 = vector

    W = np.array([[0, -v2, v1], [v2, 0, -v0], [-v1, v0, 0]])

    angle = Units.convert(angle, unit, "radian")
    a1 = np.sin(angle)
    a2 = 1.0 - np.cos(angle)

    return np.identity(3) + a1 * W + a2 * W @ W


def distance_array(array1, array2):
    """Calculates distance between each pair of points in *array1* and *array2*. Returns 2D ``numpy`` array.

    Uses fast ``cdist`` function if ``scipy`` is present, otherwise falls back to slightly slower ``numpy`` loop. Arguments should be 2-dimensional ``numpy`` arrays with the same second dimension. If *array1* is A x N and *array2* is B x N, the returned array is A x B.
    """
    return cdist(array1, array2) if scipy_present else np.array([np.linalg.norm(i - array2, axis=1) for i in array1])


def angle(vec1, vec2, result_unit="radian"):
    """Calculate an angle between vectors *vec1* and *vec2*.

    *vec1* and *vec2* should be iterable containers of length 3 (for example: tuple, list, numpy array). Values stored in them are expressed in Angstrom. Returned value is expressed in *result_unit*.

    This method requires all atomic coordinates to be numerical values, :exc:`~exceptions.TypeError` is raised otherwise.
    """
    vec1 = np.array([*vec1], dtype=float)
    vec2 = np.array([*vec2], dtype=float)

    num = np.dot(vec1, vec2)
    den = np.sqrt(((vec1) ** 2).sum()) * np.sqrt(((vec2) ** 2).sum())
    return Units.convert(np.arccos(num / den), "radian", result_unit)


def dihedral(p1, p2, p3, p4, unit="radian"):
    """Calculate the value of diherdal angle formed by points *p1*, *p2*, *p3* and *p4* in a 3D space. Arguments can be any containers with 3 numerical values, also instances of |Atom|. Returned value is always non-negative, measures the angle clockwise (looking along *p2-p3* vector) and is expressed in *unit*."""
    p1 = np.array([*p1], dtype=float)
    p2 = np.array([*p2], dtype=float)
    p3 = np.array([*p3], dtype=float)
    p4 = np.array([*p4], dtype=float)

    b0 = p1 - p2
    b1 = p3 - p2
    b2 = p4 - p3

    b1 /= np.linalg.norm(b1)
    v = b0 - np.dot(b0, b1) * b1
    w = b2 - np.dot(b2, b1) * b1

    x = np.dot(v, w)
    y = np.dot(np.cross(b1, v), w)
    ret = np.arctan2(y, x)
    ret = 2 * np.pi + ret if ret < 0 else ret
    return Units.convert(ret, "radian", unit)


def cell_shape(lattice):
    """
    Converts lattice vectors to lengths and angles (in radians)
    Sets internal cell size data, based on set of cell vectors.

    *cellvectors* is list containing three cell vectors (a 3x3 matrix)
    """
    lattice = np.asarray(lattice)
    a, b, c = np.sqrt((lattice**2).sum(axis=1))

    if a == 0.0 and b == 0.0 and c == 0.0:
        return

    alpha, beta, gamma = (90.0, 90.0, 90.0)

    if c != 0:
        alpha = angle(lattice[1], lattice[2])
        beta = angle(lattice[0], lattice[2])
    if b != 0:
        gamma = angle(lattice[0], lattice[1])

    return [a, b, c, alpha, beta, gamma]


def cell_lengths(lattice, unit="angstrom"):
    """Return the lengths of the lattice vector. Returns a list with the same length as the number of lattice vector."""

    if lattice is None or len(lattice) == 0:
        raise ValueError("Cannot calculate cell_lengths for nonperiodic system")
    lattice = np.asarray(lattice)
    ret = np.sqrt((lattice**2).sum(axis=1)) * Units.conversion_ratio("angstrom", unit)
    return ret.tolist()


def cell_angles(lattice, unit="degree"):
    """Return the angles between lattice vectors.

    unit : str
        output unit

    For 2D systems, returns a list [gamma]

    For 3D systems, returns a list [alpha, beta, gamma]
    """
    ndim = len(lattice)

    if ndim < 2:
        raise ValueError(
            "Cannot calculate cell_angles for fewer than 2 lattice vectors. Tried with {} lattice vectors".format(ndim)
        )

    gamma = angle(lattice[0], lattice[1], result_unit=unit)

    if ndim == 2:
        return [gamma]

    if ndim >= 3:
        alpha = angle(lattice[1], lattice[2], result_unit=unit)
        beta = angle(lattice[0], lattice[2], result_unit=unit)
        return [alpha, beta, gamma]


def cellvectors_from_shape(box):
    """
    Converts lengths and angles (in radians) of lattice vectors to the lattice vectors
    """
    a = box[0]
    b = box[1]
    c = box[2]
    alpha, beta, gamma = 90.0, 90.0, 90
    if len(box) == 6:
        alpha = box[3]  # *np.pi/180.
        beta = box[4]  # *np.pi/180.
        gamma = box[5]  # *np.pi/180.

    va = [a, 0.0, 0.0]
    vb = [b * np.cos(gamma), b * np.sin(gamma), 0.0]

    cx = c * np.cos(beta)
    cy = (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) * c / np.sin(gamma)
    volume = 1 - np.cos(alpha) ** 2 - np.cos(beta) ** 2 - np.cos(gamma) ** 2
    volume += 2 * np.cos(alpha) * np.cos(beta) * np.cos(gamma)
    volume = np.sqrt(volume)
    cz = c * volume / np.sin(gamma)
    vc = [cx, cy, cz]

    lattice = [va, vb, vc]

    return lattice



import os
from collections import OrderedDict

from natsort import natsorted
from scipy.optimize import curve_fit
from scm.plams.core.basejob import MultiJob
from scm.plams.core.functions import add_to_instance
from scm.plams.core.results import Results
from scm.plams.interfaces.adfsuite.ams import AMSJob, AMSResults
import numpy as np
from scm.plams.mol.molecule import Molecule
from scm.plams.recipes.md.amsmdjob import AMSNVTJob
from scm.plams.tools.kftools import KFFile

__all__ = ["AMSGreenKuboViscosityJob", "AMSGreenKuboViscosityResults", "AMSPartialGreenKuboViscosityJob"]


def get_viscosity(job, max_dt_fs=50000, reuse=False):
    """
    job: AMSGreenKuboViscosityJob
        The job

    Returns the viscosity in mPa*s, and writes

    * viscosity.txt: viscosity in mPa*s
    * green_kubo_viscosity.txt: viscosity vs. time (curve that should converge to the viscosity as t goes to inifinity)
    * fit.txt : A fit of the form A*(1-exp(-x/tau)) to green_kubo_viscosity.txt; A is the final viscosity.
    * moving_avg_visc.txt : viscosity vs. time (moving average of green_kubo_viscosity.txt)
    * info.txt : Contains the upper time-limit of the fit and the viscosity

    The upper limit for the fit is determined by when the moving average starts to decrease.

    max_dt_fs : float
        Maximum correlation time in femtoseconds

    reuse : bool
        Will return the value of viscosity.txt if it exists (and not recalculate the autocorrelation function/integral)
    """

    def f(x, A, tau):
        return A * (1 - np.exp(-x / tau))  # A is the final converged viscosity

    if reuse and os.path.exists(job.path + "/viscosity.txt"):
        return float(np.loadtxt(job.path + "/viscosity.txt"))

    AMSGreenKuboViscosityResults._accumulate_results(job.path, previous_pressuretensors=job.previous_pressuretensors)
    pressuretensor_npy = job.path + "/concatenated_pressuretensor.npy"

    pressuretensor = np.load(pressuretensor_npy)

    for child in job.children.values():
        volume = child.results.get_main_molecule().unit_cell_volume()
        break

    temperature = job.temperature
    max_dt = int(max_dt_fs / (job.samplingfreq * job.timestep))

    t, visc = AMSResults._get_green_kubo_viscosity(
        pressuretensor=pressuretensor,
        max_dt=max_dt,
        time_step=job.samplingfreq * job.timestep,
        volume=volume,
        temperature=temperature,
    )

    A = np.stack((t, visc), axis=1)
    np.savetxt(job.path + "/green_kubo_viscosity.txt", A, header="Time(fs) Viscosity(mPa*s)")

    window = min(5000, int(t[-1] / 3))
    window = max(window, 1)  # window must be greater than 0
    moving_avg_t = np.convolve(t, np.ones(window) / window, mode="valid")
    moving_avg_visc = np.convolve(visc, np.ones(window) / window, mode="valid")

    flat_region_from = min(1000, int(t[-1] / 10))  # femtoseconds
    flat_region_from_indices = moving_avg_t > flat_region_from
    allowed_t = moving_avg_t[flat_region_from_indices]
    allowed_visc = moving_avg_visc[flat_region_from_indices]
    d_moving_avg_visc = np.diff(allowed_visc)
    zero_indices = np.argwhere(np.diff(np.sign(d_moving_avg_visc))).flatten()
    if len(zero_indices) == 0:
        max_index = len(allowed_visc) - 1
    else:
        max_index = zero_indices[0]

    fit_until_t = allowed_t[max_index]

    fit_x_indices = t < fit_until_t
    fit_x = t[fit_x_indices]
    fit_y = visc[fit_x_indices]

    popt, _ = curve_fit(f, fit_x, fit_y, p0=(1.0, 1000))
    prediction = f(t, popt[0], popt[1])
    with open(job.path + "/viscosity.txt", "w") as f:
        f.write("{}\n".format(popt[0]))

    with open(job.path + "/info.txt", "w") as f:
        f.write("FITTED UNTIL t = {} fs \n".format(fit_until_t))
        f.write("VISCOSITY = {} mPa*s\n".format(popt[0]))

    A = np.stack((t, prediction), axis=1)
    np.savetxt(job.path + "/fit.txt", A, header="Time(fs) Viscosity(mPa*s)")

    A = np.stack((moving_avg_t, moving_avg_visc), axis=1)
    np.savetxt(job.path + "/moving_avg_visc.txt", A, header="Time(fs) Viscosity(mPa*s)")

    return popt[0]


class AMSPartialGreenKuboViscosityResults(AMSResults):
    def get_pressuretensor_history(self):
        pt = self.get_history_property("PressureTensor", "MDHistory")
        return np.array(pt)


class AMSPartialGreenKuboViscosityJob(AMSNVTJob):
    _result_type = AMSPartialGreenKuboViscosityResults

    def __init__(self, keep_trajectory: bool = False, max_dt_fs: float = 50000, **kwargs):
        self.keep_trajectory = keep_trajectory
        self.max_dt_fs = max_dt_fs

        kwargs.pop("calcpressure", True)
        AMSNVTJob.__init__(
            self,
            writevelocities=kwargs.get("writevelocities", False),
            writemolecules=kwargs.get("writemolecules", False),
            writebonds=kwargs.get("writebonds", False),
            writecharges=kwargs.get("writecharges", False),
            calcpressure=True,
            **kwargs,
        )

    def postrun(self):
        pt = self.results.get_pressuretensor_history()
        np.save(self.path + "/pressuretensor.npy", pt)

        if not self.keep_trajectory:
            kf = KFFile(os.path.join(self.path, "ams.rkf"))
            kf.delete_section("History")

        self.results.collect()

        get_viscosity(self.parent, max_dt_fs=self.max_dt_fs, reuse=False)


class AMSGreenKuboViscosityResults(Results):
    """Results class for AMSGreenKuboViscosityJob"""

    def get_viscosity(self, max_dt_fs=50000, reuse=False):
        """
        Returns the viscosity in mPa*s, and writes

        * viscosity.txt: viscosity in mPa*s
        * green_kubo_viscosity.txt: viscosity vs. time (curve that should converge to the viscosity as t goes to inifinity)
        * fit.txt : A fit of the form A*(1-exp(-x/tau)) to green_kubo_viscosity.txt; A is the final viscosity.
        * moving_avg_visc.txt : viscosity vs. time (moving average of green_kubo_viscosity.txt)
        * info.txt : Contains the upper time-limit of the fit and the viscosity

        The upper limit for the fit is determined by when the moving average starts to decrease.

        max_dt_fs : float
            Maximum correlation time in femtoseconds

        reuse : bool
            Will return the value of viscosity.txt if it exists (and not recalculate the autocorrelation function/integral)
        """
        return get_viscosity(self.job, max_dt_fs=max_dt_fs, reuse=reuse)

    @staticmethod
    def _accumulate_results(path=None, previous_pressuretensors=None):
        import glob

        path = path or os.getcwd()
        complete_list = []

        if previous_pressuretensors:
            complete_list.extend(previous_pressuretensors)

        for npy in natsorted(glob.glob(f"{path}/*/pressuretensor.npy")):
            complete_list.append(np.load(npy)[:-1])

        if len(complete_list) == 0:
            raise ValueError("Couldn't find any pressuretensor.npy files!")

        complete_list = np.concatenate(complete_list).reshape(-1, 6)
        np.save(path + "/concatenated_pressuretensor.npy", complete_list)
        return complete_list

    def accumulate_results(self):
        return self._accumulate_results(path=self.job.path, previous_pressuretensors=self.job.previous_pressuretensors)


class AMSGreenKuboViscosityJob(MultiJob):
    """A class for calculating the Green-Kubo viscosity of a liquid"""

    _result_type = AMSGreenKuboViscosityResults

    def __init__(
        self,
        molecule: Molecule = None,
        name: str = "greenkuboviscosity",
        nsteps: int = 100000,
        fragment_length: int = 10000,
        samplingfreq: int = 5,
        timestep: float = 1.0,
        temperature: float = 300,
        keep_trajectory: bool = False,
        restart_from=None,
        previous_pressuretensors=None,
        **kwargs,
    ):
        """
        molecule: Molecule
            3D molecule (liquid) with a low density.

        nsteps: int
            Total number of MD steps in the simulation

        fragment_length: int
            Length per "fragment simulation". Should normally not be greater than 10000, as that will slow down the writing of the trajectory.

        samplingfreq: int
            How often to sample

        timestep: float
            The timestep in femtoseconds.

        temperature: float
            The temperature in K

        keep_trajectory: bool
            Whether to keep the structures in the trajectories. This takes up a lot of disk space.

        restart_from: AMSJob
            A previous job to restart from (structure and velocities). If given, ``molecule`` is ignored. Example: ``restart_from=AMSJob.load_external('plams_workdir/greenkubo/step59')

        previous_pressuretensors: str or list of np.array
            Load previous pressure tensors. Should normally be the path to "concatenated_pressuretensors.npy" when you use ``restart_from``.

        kwargs: other options to be passed to the AMSNVTJob constructor (engine settings, ...)
        """
        MultiJob.__init__(self, children=OrderedDict(), name=name)

        self.timestep = timestep
        self.temperature = temperature
        self.samplingfreq = samplingfreq
        self.nsteps = nsteps
        self.fragment_length = fragment_length
        self.keep_trajectory = keep_trajectory
        self.restart_from = restart_from

        self.n_fragments = self.nsteps // self.fragment_length
        self.n_fragments = max(1, self.n_fragments)  # at least 1 fragment

        if self.n_fragments > 100:
            # this check is here because you otherwise run into some maximum recursion error around n_fragments = 150
            raise ValueError(
                f"AMSGreenKuboViscosityJob can be used with most 100 fragments/steps, current value: {self.n_fragments}. Either decrease nsteps (current: {self.nsteps}) or increase fragment_length (current: {self.fragment_length})."
            )

        self.previous_pressuretensors = previous_pressuretensors or []
        if isinstance(self.previous_pressuretensors, str):
            self.previous_pressuretensors = [np.load(self.previous_pressuretensors)]

        if restart_from:
            if not isinstance(restart_from, AMSJob):
                raise ValueError(f"restart_from must be AMSJob, got {type(restart_from)}")
            self.children["step1"] = AMSPartialGreenKuboViscosityJob(
                name="step1",
                keep_trajectory=self.keep_trajectory,
                timestep=self.timestep,
                temperature=self.temperature,
                nsteps=self.fragment_length,
                samplingfreq=self.samplingfreq,
                **kwargs,
            )

            @add_to_instance(self.children["step1"])
            def prerun(self):  # noqa F811
                self.get_velocities_from(restart_from, update_molecule=True)

        else:
            self.children["step1"] = AMSPartialGreenKuboViscosityJob(
                name="step1",
                molecule=molecule,
                keep_trajectory=self.keep_trajectory,
                timestep=self.timestep,
                temperature=self.temperature,
                nsteps=self.fragment_length,
                samplingfreq=self.samplingfreq,
                **kwargs,
            )

        previous_name = "step1"
        previous_names = dict()
        kwargs.pop("velocities", None)
        for i in range(2, self.n_fragments + 1):
            name = f"step{i}"
            previous_names[name] = previous_name

            job = AMSPartialGreenKuboViscosityJob(
                name=name,
                keep_trajectory=keep_trajectory,
                timestep=self.timestep,
                temperature=self.temperature,
                nsteps=fragment_length,
                samplingfreq=self.samplingfreq,
                **kwargs,
            )

            @add_to_instance(job)
            def prerun(self):  # noqa F811
                prev = previous_names[self.name]
                self.get_velocities_from(self.parent.children[prev], update_molecule=True)

            self.children[name] = job
            previous_name = name



#!/usr/bin/env amspython
from scm.plams import AMSJob, Atom, Molecule, Settings, finish, init


def main():
    init()

    mol = get_system()
    s = get_settings()

    # Note: do NOT give the job the name "quantumespresso", as that will cause file name clashes
    job = AMSJob(molecule=mol, settings=s, name="afm-FeO")
    print(job.get_input())

    job.run()

    print("Final energy (hartree)")
    print(job.results.get_energy())

    finish()


def get_settings() -> Settings:
    s = Settings()

    # s.runscript.preamble_lines = ['export SCM_QESTARTCMD="custom_start_command"']

    s.input.ams.Task = "SinglePoint"

    # Defining the Pseudopotentials
    s.input.QuantumEspresso.Pseudopotentials.Family = "SSSP-Efficiency"
    s.input.QuantumEspresso.Pseudopotentials.Functional = "PBE"

    # Use ._h and ._1 for unstructured blocks like K_Points
    s.input.QuantumEspresso.K_Points._h = "automatic"
    s.input.QuantumEspresso.K_Points._1 = "6 6 6 0 0 0"

    # Hubbard U specified in the new QE 7.1 format (different from QE 7.0)
    s.input.QuantumEspresso.Hubbard._h = "ortho-atomic"
    s.input.QuantumEspresso.Hubbard._1 = """
        U Fe1-3d 4.6
        U Fe2-3d 4.6
    """

    # When initializing many keys inside the same block
    # it can be convenient to group them like this
    s.input.QuantumEspresso.System = Settings(
        ecutwfc=40.0,
        ecutrho=240.0,
        occupations="smearing",
        smearing="gaussian",
        degauss=0.02,
        nspin=2,
        starting_magnetization=[Settings(Label="Fe1", Value=1.0), Settings(Label="Fe2", Value=-1.0)],
    )

    # You may also just use the normal PLAMS Settings dot notation
    s.input.QuantumEspresso.Electrons.conv_thr = 1.0e-8
    s.input.QuantumEspresso.Electrons.mixing_beta = 0.3

    return s


def get_system() -> Molecule:
    """

    Returns a PLAMS Molecule for FeO with the QE.Label properties set to 'Fe1'
    and 'Fe2' for the two Fe atoms

    """

    d = 4.33
    mol = Molecule()
    mol.add_atom(Atom(symbol="Fe", coords=(0, 0, 0)))
    mol.add_atom(Atom(symbol="Fe", coords=(d / 2, d / 2, 0)))
    mol.add_atom(Atom(symbol="O", coords=(0, d / 2, 0)))
    mol.add_atom(Atom(symbol="O", coords=(d / 2, 0, 0)))

    mol.lattice = [[d, 0, 0], [0, d, 0], [d / 2, 0, d / 2]]

    mol[1].properties.QE.Label = "Fe1"
    mol[2].properties.QE.Label = "Fe2"

    return mol


if __name__ == "__main__":
    main()



from scm.plams.core.basejob import SingleJob
from scm.plams.interfaces.adfsuite.scmjob import SCMResults

__all__ = ["MOPACJob", "MOPACResults"]


class MOPACResults(SCMResults):
    """A class for results of computation done with MOPAC.

    This class inherits all methods from |SCMResults|.

    .. technical::

        In case of a MOPAC job, preparation is much different from other programs of AMSuite, but the result handling is quite similar due to presence of KF files. Therefore |MOPACResults| is a subclass of |SCMResults|, but |MOPACJob| is not a subclass of |SCMJob|.
    """

    _rename_map = {"results.rkf": "$JN.rkf", "$JN.in.aux": "$JN.aux", "$JN.in.arc": "$JN.arc", "$JN.in.out": "$JN.out"}
    _kfext = ".rkf"

    def _int2inp(self):
        return list(range(1, 1 + len(self.job.molecule)))

    def _atomic_numbers_input_order(self):
        return self.readkf("Molecule", "AtomicNumbers")


class MOPACJob(SingleJob):
    """A class representing a single computational job with MOPAC."""

    _result_type = MOPACResults
    _command = "MOPAC2016-SCM.exe"

    def get_input(self):
        """Transform the contents of ``input`` branch of ``settings`` into the first line of MOPAC input. Print the molecular coordinates together with frozen coordinate flags."""
        aux = self.settings.input.find_case("aux")
        if aux not in self.settings.input:
            self.settings.input[aux] = [0, "PRECISION=9"]

        keylist = []
        for key, value in self.settings.input.items():
            if value is True:
                keylist.append(key)
            elif isinstance(value, tuple):
                keylist.append(key + "=(" + ",".join(map(str, value)) + ")")
            elif isinstance(value, list):
                keylist.append(key + "(" + ",".join(map(str, value)) + ")")
            else:
                keylist.append(key + "=" + str(value))
        ret = " ".join(keylist) + "\n\n\n"

        for at in self.molecule:
            line = "{:7}".format(at.symbol)
            for c in ["x", "y", "z"]:
                num = at.__getattribute__(c)
                frz = 0 if ("mopac_freeze" in at.properties and c in at.properties.mopac_freeze) else 1
                line += " {: 11.8f} {:d}".format(num, frz)
            ret += line + "\n"
        return ret

    def get_runscript(self):
        """Generate a MOPAC runscript.

        The name of the MOPAC executable is taken from class attribute ``MOPACJob._command``. If you experience problems running MOPAC, check if that value corresponds to the name of the executable and this executable is visible in your ``$PATH`` (in case of AMSuite it's in ``$AMSBIN``). Note that a bare MOPAC executable should be used here, please avoid using any wrappers.

        The execution of MOPAC binary is followed by calling a simple command line tool ``tokf`` which reads various output text files produced by MOPAC and collects all the data in a binary KF file. See :ref:`kf-files` for details.
        """
        ret = self._command + " " + self._filename("inp")
        if self.settings.runscript.stdout_redirect:
            ret += " >" + self._filename("out")
        ret += "\n\n"
        ret += "cp {} {}.stdout\n".format(self._filename("err"), self._filename("inp"))
        ret += "tokf mopac {} {}.rkf\n".format(self._filename("inp"), self.name)
        ret += "rm {}.stdout\n\n".format(self._filename("inp"))
        ret += "rm {}\n\n".format(self._filename("out"))
        return ret

    def check(self):
        """Grep standard output for ``* JOB ENDED NORMALLY *``."""
        s = self.results.grep_output("* JOB ENDED NORMALLY *")
        return len(s) > 0



import os
import shutil
from collections import OrderedDict

from scm.plams.core.basejob import MultiJob
from scm.plams.core.functions import add_to_instance
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.interfaces.adfsuite.crs import CRSJob, CRSResults
from scm.plams.mol.molecule import Molecule
from scm.plams.recipes.adfcosmorscompound import ADFCOSMORSCompoundJob
from scm.plams.tools.units import Units

"""
    Classes for calculating oxidation potentials and reduction potentials in implicit solvent.

    To get the results, call

    job.results.get_oxidation_potential(vibrations=False)

    or

    job.results.get_reduction_potential(vibrations=False)

    Set ``vibrations=True`` to include vibrational effects if the job was run
    with ``vibrations=True``.

    The potential is returned in V. It is on an absolute scale. To get
    potential relative to SHE, subtract 4.47 V.

    AMSRedoxDirectJob
    --------------------

    Calculates reaction energies for 
    A + e^- --> A^-     if ``reduction=True``,
    and/or
    A --> A^+ + e^-     if ``oxidation=True``.

    using direct geometry optimizations in implicit solvent. The solvent must be one supported by ADF.

    If ``vibrations`` = True, then calculate and use Gibbs free energies at room temperature.

    ``settings`` should be set to the engine settings, excluding spin polarization and implicit solvation.

    Requires an ADF license.

    AMSRedoxThermodynamicCycleJob
    -----------------------------

    Sometimes more efficient (and less accurate) alternative to AMSRedoxDirectJob. Can be
    useful if the molecule is large and if ``vibrations=True``. For small molecules just use AMSRedoxDirectJob.

    Requires an ADF license.

    AMSRedoxScreeningJob
    -----------------------

    The fastest and least accurate option. Geometries are optimized at the DFTB
    level, ADF is then used to generate a .coskf file and the solvation free
    energies are evaluated with COSMO-RS.

    Note: you must supply ``solvent_coskf``, a path to the solvent .coskf file.
    If you do not have one, you can generate one with the
    ``ADFCOSMORSCompoundJob`` recipe.

    Requires DFTB, ADF, and COSMO-RS licenses.

    Note: The functional used is different from Belic et al. This class uses the
    settings that are meant to be used for generating .coskf files.

"""

__all__ = [
    "AMSRedoxScreeningJob",
    "AMSRedoxScreeningResults",
    "AMSRedoxDirectJob",
    "AMSRedoxDirectResults",
    "AMSRedoxThermodynamicCycleJob",
    "AMSRedoxThermodynamicCycleResults",
]


def _spinpol_settings(molecule, charge=0):
    sett = Settings()
    num_electrons = sum(atom.atnum for atom in molecule) - charge
    spinpol = num_electrons % 2
    if spinpol == 0:
        unrestricted = "No"
    else:
        unrestricted = "Yes"
        # sett.input.adf.OCCUPATIONS = "ElectronicTemperature=100"
    sett.input.adf.SpinPolarization = spinpol
    sett.input.adf.Unrestricted = unrestricted
    return sett


class AMSRedoxParentJob(MultiJob):
    def __init__(
        self,
        molecule: Molecule,  # Molecule
        name: str = None,
        settings: Settings = None,
        oxidation: bool = True,
        reduction: bool = False,
    ):

        MultiJob.__init__(self, children=OrderedDict(), name=name)

        self.oxidation = oxidation
        self.reduction = reduction
        self.input_molecule = molecule
        self.settings = settings or Settings()
        self.orig_charge = molecule.properties.get("charge", 0)
        self.ox_charge = self.orig_charge + 1
        self.red_charge = self.orig_charge - 1

    @staticmethod
    def solvation_settings(solvent):
        sett = Settings()
        if solvent is None:
            solvent = "vacuum"
        if isinstance(solvent, str):
            sett.input.adf.Solvation.Solv = f"name={solvent}"
        elif isinstance(solvent, tuple):
            sett.input.adf.Solvation.Solv = f"eps={solvent[0]} rad={solvent[1]}"
        else:
            raise TypeError(f"solvent is of type {type(solvent)}, expected str or 2-tuple")

        return sett

    def get_dft_0_settings(self, vibrations=False, perturbcoordinates=False):
        """all settings except task and solvation"""
        s = Settings()
        s += self.settings
        s.update(_spinpol_settings(self.input_molecule, self.orig_charge))
        s.input.ams.System.Charge = self.orig_charge
        s.input.ams.System.PerturbCoordinates = 0.01 if perturbcoordinates else 0.00
        s.input.ams.Properties.NormalModes = str(vibrations)
        return s

    def get_dft_ox_settings(self, vibrations=False, perturbcoordinates=False):
        sox = self.get_dft_0_settings(vibrations=vibrations, perturbcoordinates=perturbcoordinates)
        sox.update(_spinpol_settings(self.input_molecule, self.ox_charge))
        sox.input.ams.System.Charge = self.ox_charge
        return sox

    def get_dft_red_settings(self, vibrations=False, perturbcoordinates=False):
        sred = self.get_dft_0_settings(vibrations=vibrations, perturbcoordinates=perturbcoordinates)
        sred.update(_spinpol_settings(self.input_molecule, self.red_charge))
        sred.input.ams.System.Charge = self.red_charge
        return sred


class AMSRedoxParentResults(Results):
    Gelectron = -0.0375 / 27.211  # in hartree

    @staticmethod
    def _get_energy(job, vibrations: bool):
        if not vibrations:
            return job.results.get_energy()
        e = job.results.readrkf("Thermodynamics", "Gibbs free Energy", file="engine")
        if isinstance(e, list):
            return e[0]
        return e

    @staticmethod
    def _get_solvation_energy(job):
        dG_solvation = job.results.readrkf("Energy", "Solvation Energy (el)", "adf") + job.results.readrkf(
            "Energy", "Solvation Energy (cd)", "adf"
        )
        return dG_solvation


class CRSActivityCoefficientResults(CRSResults):
    def get_gibbs_energy(self):
        """Gibbs energy in hartree"""
        return float(self.get_results()["G solute"][1]) * 0.00159376  # kcal/mol to hartree


class CRSActivityCoefficientJob(CRSJob):
    _result_type = CRSActivityCoefficientResults

    def __init__(self, solvent_coskf, solute_coskf, name=None, temperature=298.15, copy_coskf=False):
        CRSJob.__init__(self, name=name)
        self.solvent_coskf = solvent_coskf
        self.solute_coskf = solute_coskf
        self.temperature = temperature
        self.copy_coskf = copy_coskf

    def prerun(self):  # noqa F811
        self._prerun()

    def _prerun(self):
        if self.copy_coskf:
            new_solvent = f"solvent_{os.path.basename(self.solvent_coskf)}"
            new_solute = f"solute_{os.path.basename(self.solute_coskf)}"
            shutil.copy(self.solvent_coskf, os.path.join(self.path, new_solvent))
            shutil.copy(self.solute_coskf, os.path.join(self.path, new_solute))
            self.solvent_coskf = new_solvent
            self.solute_coskf = new_solute
        self.settings.input.property._h = "ACTIVITYCOEF"
        compounds = [Settings(), Settings()]
        compounds[0]._h = self.solvent_coskf
        compounds[1]._h = self.solute_coskf
        compounds[0].frac1 = 1
        compounds[1].frac1 = 0

        self.settings.input.temperature = str(self.temperature)
        self.settings.input.compound = compounds


class AMSRedoxDirectResults(AMSRedoxParentResults):
    def get_oxidation_potential(self, vibrations=True, unit="eV"):
        Greact = self._get_energy(self.job.children["job_0"], vibrations=vibrations)
        Gprod = self._get_energy(self.job.children["job_ox"], vibrations=vibrations) + self.Gelectron
        ret = (Gprod - Greact) * Units.convert(1.0, "hartree", unit)
        return ret

    def get_reduction_potential(self, vibrations=True, unit="eV"):
        Greact = self._get_energy(self.job.children["job_0"], vibrations=vibrations) + self.Gelectron
        Gprod = self._get_energy(self.job.children["job_red"], vibrations=vibrations)
        ret = (Gprod - Greact) * Units.convert(1.0, "hartree", unit)
        ret *= -1  # deltaG = -nFE
        return ret


class AMSRedoxDirectJob(AMSRedoxParentJob):
    _result_type = AMSRedoxDirectResults

    def __init__(
        self,
        molecule: Molecule,  # Molecule
        solvent: str,  # ADF pre-defined solvent
        name: str = None,
        settings: Settings = None,
        oxidation: bool = True,
        reduction: bool = False,
        vibrations: bool = True,
    ):

        AMSRedoxParentJob.__init__(
            self, name=name, settings=settings, molecule=molecule, oxidation=oxidation, reduction=reduction
        )

        self.solvent = solvent
        self.vibrations = vibrations

        s = Settings()
        s += self.settings
        s.input.ams.Task = "GeometryOptimization"
        s += self.solvation_settings(self.solvent)
        s.update(_spinpol_settings(self.input_molecule, self.orig_charge))
        s.input.ams.System.Charge = self.orig_charge
        s.input.ams.System.PerturbCoordinates = 0.01
        s.input.ams.Properties.NormalModes = str(self.vibrations)
        job_0 = AMSJob(settings=s, name="job_0", molecule=self.input_molecule)
        self.children["job_0"] = job_0

        if oxidation:
            sox = s.copy()
            sox.update(_spinpol_settings(self.input_molecule, self.ox_charge))
            sox.input.ams.System.Charge = self.ox_charge
            sox.input.ams.System.PerturbCoordinates = 0.01
            job_ox = AMSJob(settings=sox, name="job_ox")

            @add_to_instance(job_ox)
            def prerun(self):  # noqa F811
                self.molecule = job_0.results.get_main_molecule()

            self.children["job_ox"] = job_ox

        if reduction:
            sred = s.copy()
            sred.update(_spinpol_settings(self.input_molecule, self.red_charge))
            sred.input.ams.System.Charge = self.red_charge
            sred.input.ams.System.PerturbCoordinates = 0.01
            job_red = AMSJob(settings=sred, name="job_red")

            @add_to_instance(job_red)
            def prerun(self):  # noqa F811
                self.molecule = job_0.results.get_main_molecule()

            self.children["job_red"] = job_red


class AMSRedoxScreeningResults(Results):
    Gelectron = -0.0375 / 27.211  # in hartree

    def get_oxidation_potential(self, vibrations=True, unit="eV"):
        """Note: vibrations cannot be disabled for AMSRedoxScreening"""
        Greact = self.job.children["activitycoef_0"].results.get_gibbs_energy()
        Gprod = self.job.children["activitycoef_ox"].results.get_gibbs_energy() + self.Gelectron
        ret = (Gprod - Greact) * Units.convert(1.0, "hartree", unit)
        return ret

    def get_reduction_potential(self, vibrations=True, unit="eV"):
        """Note: vibrations cannot be disabled for AMSRedoxScreening"""
        Greact = self.job.children["activitycoef_0"].results.get_gibbs_energy() + self.Gelectron
        Gprod = self.job.children["activitycoef_red"].results.get_gibbs_energy()
        ret = (Gprod - Greact) * Units.convert(1.0, "hartree", unit)
        ret *= -1  # deltaG = -nFE
        return ret


class AMSRedoxScreeningJob(AMSRedoxParentJob):
    _result_type = AMSRedoxScreeningResults

    def __init__(
        self,
        molecule: Molecule,  # Molecule
        solvent_coskf: str,  # path to solvent .coskf file
        name: str = None,
        settings: Settings = None,
        oxidation: bool = True,
        reduction: bool = False,
        copy_coskf=False,
    ):
        """
        copy_coskf: bool
            Will copy the coskf files into the job's own directory before running.
        """
        AMSRedoxParentJob.__init__(
            self, name=name, settings=settings, molecule=molecule, oxidation=oxidation, reduction=reduction
        )

        self.solvent_coskf = solvent_coskf

        # dftb_go_0
        s = Settings()
        s.input.ams.Task = "GeometryOptimization"
        s.input.DFTB.Model = "GFN1-xTB"
        s.input.ams.System.Charge = self.orig_charge
        s.input.ams.System.PerturbCoordinates = 0.01
        dftb_go_0 = AMSJob(settings=s, molecule=self.input_molecule, name="dftb_go_0")
        self.children["dftb_go_0"] = dftb_go_0

        if oxidation:
            sox = s.copy()
            sox.input.ams.System.PerturbCoordinates = 0.01
            sox.input.ams.System.Charge = self.ox_charge
            dftb_go_ox = AMSJob(settings=sox, name="dftb_go_ox")

            @add_to_instance(dftb_go_ox)
            def prerun(self):  # noqa F811
                self.molecule = dftb_go_0.results.get_main_molecule()

            self.children["dftb_go_ox"] = dftb_go_ox

        if reduction:
            sred = s.copy()
            sred.input.ams.System.PerturbCoordinates = 0.01
            sred.input.ams.System.Charge = self.red_charge
            dftb_go_red = AMSJob(settings=sred, name="dftb_go_red")

            @add_to_instance(dftb_go_red)
            def prerun(self):  # noqa F811
                self.molecule = dftb_go_0.results.get_main_molecule()

            self.children["dftb_go_red"] = dftb_go_red

        # gencoskf_0
        s = self.settings.copy()
        s.update(_spinpol_settings(self.input_molecule, self.orig_charge))
        # s.input.ams.System.Charge = self.orig_charge
        gencoskf_0 = ADFCOSMORSCompoundJob(molecule=None, name="gencoskf_0", singlepoint=True, settings=s)

        @add_to_instance(gencoskf_0)
        def prerun(self):  # noqa F811
            self.input_molecule = dftb_go_0.results.get_main_molecule()  # will inherit the charge
            assert isinstance(self.input_molecule, Molecule)
            self.atomic_ion = len(self.input_molecule) == 1

        self.children["gencoskf_0"] = gencoskf_0

        if oxidation:
            sox = s.copy()
            sox.update(_spinpol_settings(self.input_molecule, self.ox_charge))
            # sox.input.ams.System.Charge = self.ox_charge
            gencoskf_ox = ADFCOSMORSCompoundJob(molecule=None, name="gencoskf_ox", singlepoint=True, settings=sox)

            @add_to_instance(gencoskf_ox)
            def prerun(self):  # noqa F811
                self.input_molecule = dftb_go_ox.results.get_main_molecule()
                self.atomic_ion = len(self.input_molecule) == 1

            self.children["gencoskf_ox"] = gencoskf_ox

        if reduction:
            sred = s.copy()
            sred.update(_spinpol_settings(self.input_molecule, self.red_charge))
            # sred.input.ams.System.Charge = self.red_charge
            gencoskf_red = ADFCOSMORSCompoundJob(molecule=None, name="gencoskf_red", singlepoint=True, settings=sred)

            @add_to_instance(gencoskf_red)
            def prerun(self):  # noqa F811
                self.input_molecule = dftb_go_red.results.get_main_molecule()
                self.atomic_ion = len(self.input_molecule) == 1

            self.children["gencoskf_red"] = gencoskf_red

        # activitycoef_0
        activitycoef_0 = CRSActivityCoefficientJob(
            name="activitycoef_0", solvent_coskf=self.solvent_coskf, solute_coskf=None, copy_coskf=copy_coskf
        )

        @add_to_instance(activitycoef_0)
        def prerun(self):  # noqa F811
            self.solute_coskf = gencoskf_0.results.coskfpath()
            self._prerun()

        self.children["activitycoef_0"] = activitycoef_0

        if oxidation:
            activitycoef_ox = CRSActivityCoefficientJob(
                name="activitycoef_ox", solvent_coskf=self.solvent_coskf, solute_coskf=None, copy_coskf=copy_coskf
            )

            @add_to_instance(activitycoef_ox)
            def prerun(self):  # noqa F811
                self.solute_coskf = gencoskf_ox.results.coskfpath()
                self._prerun()

            self.children["activitycoef_ox"] = activitycoef_ox

        if reduction:
            activitycoef_red = CRSActivityCoefficientJob(
                name="activitycoef_red", solvent_coskf=self.solvent_coskf, solute_coskf=None, copy_coskf=copy_coskf
            )

            @add_to_instance(activitycoef_red)
            def prerun(self):  # noqa F811
                self.solute_coskf = gencoskf_red.results.coskfpath()
                self._prerun()

            self.children["activitycoef_red"] = activitycoef_red


class AMSRedoxThermodynamicCycleResults(AMSRedoxParentResults):
    def get_oxidation_potential(self, vibrations=True, unit="eV"):
        Greact = (
            self._get_energy(self.job.children["go_0_vacuum"], vibrations=vibrations)
            + self._get_energy(self.job.children["go_0_vacuum_sp_solvated"], vibrations=False)
            + self._get_energy(self.job.children["go_0_solvated_sp_vacuum"], vibrations=False)
            - 2 * self._get_energy(self.job.children["go_0_vacuum"], vibrations=False)
        )

        Gprod = (
            self._get_energy(self.job.children["go_ox_vacuum"], vibrations=vibrations)
            + self._get_energy(self.job.children["go_ox_vacuum_sp_solvated"], vibrations=False)
            + self._get_energy(self.job.children["go_ox_solvated_sp_vacuum"], vibrations=False)
            - 2 * self._get_energy(self.job.children["go_ox_vacuum"], vibrations=False)
            + self.Gelectron
        )

        ret = (Gprod - Greact) * Units.convert(1.0, "hartree", unit)
        return ret

    def get_reduction_potential(self, vibrations=True, unit="eV"):
        Greact = (
            self._get_energy(self.job.children["go_0_vacuum"], vibrations=vibrations)
            + self._get_energy(self.job.children["go_0_vacuum_sp_solvated"], vibrations=False)
            + self._get_energy(self.job.children["go_0_solvated_sp_vacuum"], vibrations=False)
            - 2 * self._get_energy(self.job.children["go_0_vacuum"], vibrations=False)
            + self.Gelectron
        )

        Gprod = (
            self._get_energy(self.job.children["go_red_vacuum"], vibrations=vibrations)
            + self._get_energy(self.job.children["go_red_vacuum_sp_solvated"], vibrations=False)
            + self._get_energy(self.job.children["go_red_solvated_sp_vacuum"], vibrations=False)
            - 2 * self._get_energy(self.job.children["go_red_vacuum"], vibrations=False)
        )

        ret = (Gprod - Greact) * Units.convert(1.0, "hartree", unit)
        ret *= -1  # deltaG = -nFE
        return ret


class AMSRedoxThermodynamicCycleJob(AMSRedoxParentJob):
    _result_type = AMSRedoxThermodynamicCycleResults

    def __init__(
        self,
        molecule: Molecule,  # Molecule
        name: str = None,
        solvent: str = "Water",
        settings: Settings = None,
        oxidation: bool = True,
        reduction: bool = False,
        vibrations: bool = False,
    ):

        AMSRedoxParentJob.__init__(
            self, name=name, settings=settings, molecule=molecule, oxidation=oxidation, reduction=reduction
        )

        self.vibrations = vibrations
        self.solvent = solvent

        # go_0_vacuum
        s = self.get_dft_0_settings(vibrations=self.vibrations, perturbcoordinates=True)
        s.input.ams.Task = "GeometryOptimization"
        go_0_vacuum = AMSJob(settings=s, molecule=self.input_molecule, name="go_0_vacuum")

        @add_to_instance(go_0_vacuum)
        def prerun(self):  # noqa F811
            self.molecule = self.parent.input_molecule

        self.children["go_0_vacuum"] = go_0_vacuum

        # go_0_vacuum_sp_solvated
        ss = self.get_dft_0_settings(vibrations=False, perturbcoordinates=False)
        ss.input.ams.Task = "SinglePoint"
        ss += self.solvation_settings(self.solvent)
        go_0_vacuum_sp_solvated = AMSJob(settings=ss, molecule=None, name="go_0_vacuum_sp_solvated")

        @add_to_instance(go_0_vacuum_sp_solvated)
        def prerun(self):  # noqa F811
            self.molecule = go_0_vacuum.results.get_main_molecule()

        self.children["go_0_vacuum_sp_solvated"] = go_0_vacuum_sp_solvated

        # go_0_solvated
        s = self.get_dft_0_settings(vibrations=self.vibrations, perturbcoordinates=True)
        s.input.ams.Task = "GeometryOptimization"
        s += self.solvation_settings(self.solvent)
        go_0_solvated = AMSJob(settings=s, molecule=None, name="go_0_solvated")

        @add_to_instance(go_0_solvated)
        def prerun(self):  # noqa F811
            self.molecule = go_0_vacuum.results.get_main_molecule()

        self.children["go_0_solvated"] = go_0_solvated

        # go_0_solvated_sp_vacuum
        ss = self.get_dft_0_settings(vibrations=False, perturbcoordinates=False)
        ss.input.ams.Task = "SinglePoint"
        go_0_solvated_sp_vacuum = AMSJob(settings=ss, molecule=None, name="go_0_solvated_sp_vacuum")

        @add_to_instance(go_0_solvated_sp_vacuum)
        def prerun(self):  # noqa F811
            self.molecule = go_0_solvated.results.get_main_molecule()

        self.children["go_0_solvated_sp_vacuum"] = go_0_solvated_sp_vacuum

        if oxidation:
            # go_ox_vacuum
            s = self.get_dft_ox_settings(vibrations=self.vibrations, perturbcoordinates=True)
            s.input.ams.Task = "GeometryOptimization"
            go_ox_vacuum = AMSJob(settings=s, molecule=self.input_molecule, name="go_ox_vacuum")

            @add_to_instance(go_ox_vacuum)
            def prerun(self):  # noqa F811
                self.molecule = go_0_vacuum.results.get_main_molecule()

            self.children["go_ox_vacuum"] = go_ox_vacuum

            # go_ox_vacuum_sp_solvated
            ss = self.get_dft_ox_settings(vibrations=False, perturbcoordinates=False)
            ss.input.ams.Task = "SinglePoint"
            ss += self.solvation_settings(self.solvent)
            go_ox_vacuum_sp_solvated = AMSJob(settings=ss, molecule=None, name="go_ox_vacuum_sp_solvated")

            @add_to_instance(go_ox_vacuum_sp_solvated)
            def prerun(self):  # noqa F811
                self.molecule = go_ox_vacuum.results.get_main_molecule()

            self.children["go_ox_vacuum_sp_solvated"] = go_ox_vacuum_sp_solvated

            # go_ox_solvated
            s = self.get_dft_ox_settings(vibrations=self.vibrations, perturbcoordinates=True)
            s.input.ams.Task = "GeometryOptimization"
            s += self.solvation_settings(self.solvent)
            go_ox_solvated = AMSJob(settings=s, molecule=None, name="go_ox_solvated")

            @add_to_instance(go_ox_solvated)
            def prerun(self):  # noqa F811
                self.molecule = go_ox_vacuum.results.get_main_molecule()

            self.children["go_ox_solvated"] = go_ox_solvated

            # go_ox_solvated_sp_vacuum
            ss = self.get_dft_ox_settings(vibrations=False, perturbcoordinates=False)
            ss.input.ams.Task = "SinglePoint"
            go_ox_solvated_sp_vacuum = AMSJob(settings=ss, molecule=None, name="go_ox_solvated_sp_vacuum")

            @add_to_instance(go_ox_solvated_sp_vacuum)
            def prerun(self):  # noqa F811
                self.molecule = go_ox_solvated.results.get_main_molecule()

            self.children["go_ox_solvated_sp_vacuum"] = go_ox_solvated_sp_vacuum

        if reduction:
            # go_red_vacuum
            s = self.get_dft_red_settings(vibrations=self.vibrations, perturbcoordinates=True)
            s.input.ams.Task = "GeometryOptimization"
            go_red_vacuum = AMSJob(settings=s, molecule=self.input_molecule, name="go_red_vacuum")

            @add_to_instance(go_red_vacuum)
            def prerun(self):  # noqa F811
                self.molecule = go_0_vacuum.results.get_main_molecule()

            self.children["go_red_vacuum"] = go_red_vacuum

            # go_red_vacuum_sp_solvated
            ss = self.get_dft_red_settings(vibrations=False, perturbcoordinates=False)
            ss.input.ams.Task = "SinglePoint"
            ss += self.solvation_settings(self.solvent)
            go_red_vacuum_sp_solvated = AMSJob(settings=ss, molecule=None, name="go_red_vacuum_sp_solvated")

            @add_to_instance(go_red_vacuum_sp_solvated)
            def prerun(self):  # noqa F811
                self.molecule = go_red_vacuum.results.get_main_molecule()

            self.children["go_red_vacuum_sp_solvated"] = go_red_vacuum_sp_solvated

            # go_red_solvated
            s = self.get_dft_red_settings(vibrations=self.vibrations, perturbcoordinates=True)
            s.input.ams.Task = "GeometryOptimization"
            s += self.solvation_settings(self.solvent)
            go_red_solvated = AMSJob(settings=s, molecule=None, name="go_red_solvated")

            @add_to_instance(go_red_solvated)
            def prerun(self):  # noqa F811
                self.molecule = go_red_vacuum.results.get_main_molecule()

            self.children["go_red_solvated"] = go_red_solvated

            # go_red_solvated_sp_vacuum
            ss = self.get_dft_red_settings(vibrations=False, perturbcoordinates=False)
            ss.input.ams.Task = "SinglePoint"
            go_red_solvated_sp_vacuum = AMSJob(settings=ss, molecule=None, name="go_red_solvated_sp_vacuum")

            @add_to_instance(go_red_solvated_sp_vacuum)
            def prerun(self):  # noqa F811
                self.molecule = go_red_solvated.results.get_main_molecule()

            self.children["go_red_solvated_sp_vacuum"] = go_red_solvated_sp_vacuum
